/*

* Copyright (c) 2005, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

import java.io.*;
import java.util.*;

/**
 * Thread accepts sensor queries and returns simulated query results.
 *
 * @see ScanOperator 
 */
public class AortaSimulatedSensorQuery extends Thread {
	
	/** Scan operator this simulated sensor query outputs sensor tuple to. */
	public ScanOperator scanOp;
	
	/** SQL text of this simulated sensor query. */
	public String sensorQueryString;
		
	int attrNum = 1;
	
	/** Total number of sensors in the network. */
	public static final int sensorNum = 10;
	
	public static final int sleep_interval_min = 30; // unit: second
	public static final int sleep_interval_max = 60;
	
	String attributes [] = new String [50];
	static final int attr_value_min = 400;
	static final int attr_value_max = 500;
	
	boolean queryStop = false;
	
	int epoch = 1;
	
	public AortaSimulatedSensorQuery(String sensorQueryString, ScanOperator scanOp)
	  {
	  	try
	  	  {
	  	  	this.scanOp = scanOp;
	  	  	
	  	  	this.sensorQueryString = sensorQueryString;	  	
	  	  	
	  	  	int start,stop,end;
	  	  	
	  	  	start = sensorQueryString.indexOf("select ");
	  	  	start += 7;
	  	  	
	  	  	end = sensorQueryString.indexOf(" from ");
	  	  	
	  	  	// decide how many attributes involved in the SELECT clause
	  	  	start = sensorQueryString.indexOf(",",start);
	  	  	while ( (start != -1) && (start < end) )
	  	  	  {
	  	  	  	attrNum ++;
	  	  	    start = start + 1;
	  	  	    stop = sensorQueryString.indexOf(",",start);
	  	  	    if (stop != -1)
	  	  	      attributes[attrNum - 1] = sensorQueryString.substring(start,stop).trim();
	  	  	    else 
	  	  	      attributes[attrNum - 1] = sensorQueryString.substring(start,end).trim();
	  	  	    start = stop;
	  	      }
	  	    
	  	  }
	  	catch (Exception e) {e.printStackTrace();}	  	
	  	
	  }
	  
	
	/** 
	 * Simulate the execution of a query over a sensor network.
	 * Output a simulated sensor tuple to {@link #scanOp} every few seconds. 
	 */
	public void run()
	  {
	  	Random ran = new Random ();
	    long sleep_interval = 0;
	    String str;
	    int sid, attrV;	    
	  	
	  	while (!queryStop)
	  	  {
	  	  	sleep_interval = sleep_interval_min + Math.abs(ran.nextLong()) % (sleep_interval_max - sleep_interval_min + 1);
	  	    sleep_interval *= 1000;
	  	    sleep_interval += Aorta.DEFAULT_SAMPLE_PERIOD;
	  	    
	  	    try {Thread.currentThread().sleep(sleep_interval);}
	  	  	catch (Exception e) {e.printStackTrace();}
	  	  	
	  	  	scanOp.q.logFileWriter.println("An event occurs!");
	  	  	scanOp.q.eventStartTime = System.currentTimeMillis();
	  	  	
	  	  	scanOp.q.logFileWriter.print("Simulated sensor tuple: ");
	  	  	
	  	  	Vector v = new Vector(10,1);
	  	  	sid = Math.abs(ran.nextInt()) % sensorNum + 1;
	  	  	if (sid >= 5) 
	  	  	  sid = Math.abs(ran.nextInt()) % sensorNum + 1;
	  	  	if (sid >= 3)
	  	  	  sid = Math.abs(ran.nextInt()) % sensorNum + 1;	  	  	
	  	  	  
	  	  	scanOp.q.logFileWriter.print("epoch = " + epoch + " ");
	  	  	scanOp.q.logFileWriter.print("nodeid = " + sid + " ");
	  	  	scanOp.q.logFileWriter.println();
	  	  	// System.out.println("\nTriggering sensor id: " + sid + "\tQuery id: " + scanOp.q.id); 
	  	  	v.addElement(String.valueOf(epoch));
	  	  	epoch ++;
	  	  	v.addElement(String.valueOf(sid));
	  	  	
	  	  	for (int i=1;i<attrNum;i++)
	  	  	  {
	  	  	  	attrV = attr_value_min + Math.abs(ran.nextInt()) % (attr_value_max - attr_value_min + 1);
	  	  	  	v.addElement(String.valueOf(attrV));
	  	  	  	scanOp.q.logFileWriter.print(attributes[i] + "=" + attrV + " ");
	  	  	  }
	  	  	
	  	  	
	  	  	scanOp.q.logFileWriter.println("Sensor query string: " + sensorQueryString);
	  	  	scanOp.q.logFileWriter.println("Sleep interval: " + sleep_interval + " ms.");
	  	  	scanOp.q.logFileWriter.println();
	  	  	scanOp.q.logFileWriter.flush();
	  	  	
	  	  	Tuple t = new Tuple (v);	  	  	
	  	  	
	  	  	scanOp.inputQueue.addElement((Tuple)t);  
	  	  	scanOp.getInput();	        	  	  		  	  	
	  	  }
	  }
	
	
   /** Stop this simulated sensor query. */
   public void stopSimulatedSensorQuery ()
	  {
	  	queryStop = true;  	
	  	
	  	System.gc();
	  }
	
}