/*

* Copyright (c) 2005, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

import java.io.*;
import java.util.*;

/**
 * Implementation of Algorithm 1.1 (LERFA: Least Eligible Request First Assignment).
 *
 * @see SRFE
 * @see AortaSchedulingSimulator#schedule
 * @see GroupingThread#scheduleCameras
 */
public class LERFA {
	
  Vector requests;
  
  Workload deviceWorkload [];
	
  public LERFA (Vector requests)
    {
        this.requests = requests;    	    	
    	    	
    	Vector didVector = new Vector (20,10);
    	int i,j;
    	for (i=0;i<requests.size();i++)
    	  for (j=0;j<((ActionRequest)requests.elementAt(i)).didVector.size();j++)
    	  	if (!didVector.contains((String)((ActionRequest)requests.elementAt(i)).didVector.elementAt(j)))
    	  	  didVector.addElement(((String)((ActionRequest)requests.elementAt(i)).didVector.elementAt(j)));
    	deviceWorkload = new Workload [didVector.size()];
    	
    	for (i=0;i<didVector.size();i++)
    	  {
    	  	deviceWorkload[i] = new Workload ();
    	  	deviceWorkload[i].did = (String)(didVector.elementAt(i));
    	  	deviceWorkload[i].workload = 0;    	  	
    	  }    	
    } 
    
    
    /** Assign the requests on the canidate devices with a least eligible request first heurisic. */
    public void assignment ()
      {
      	int size = requests.size();
      	int count = 0;
      	int sel_no;
      	int i,j,k;
      	double workload, cost;
      	String did;
      	ActionRequest ar;
      	Vector sorted_requests = new Vector (20,10);
      	                
      	
        for (i=1;;i++)
          {
          	for (j=0;j<requests.size();j++)
          	  if (((ActionRequest)requests.elementAt(j)).didVector.size() == i)
          	  	{
          	  	  sorted_requests.addElement((ActionRequest)requests.elementAt(j));
          	  	  count ++;
          	    }
          	 
          	 if (count == size) 
          	   break;
          } 
       
        for (i=0;i<size;i++)
          {
            ar = (ActionRequest)sorted_requests.elementAt(i);
            
            workload = Double.MAX_VALUE;
            
            sel_no = -1;
            
         	for (j=0;j<ar.didVector.size();j++)
         	  {
         	  	
         	  	did = (String)ar.didVector.elementAt(j);
         	  	cost = Double.parseDouble((String)ar.costVector.elementAt(j));
         	  	
         	  	for (k=0;k<deviceWorkload.length;k++)
         	  	  if (did.equals(deviceWorkload[k].did))
         	  	    break;
         	  	    
         	  	if (cost + deviceWorkload[k].workload < workload)
         	  	  {
         	  	  	workload = cost + deviceWorkload[k].workload;
         	  	  	sel_no = k;
         	  	  	ar.sel_no = j;         	  	  	
         	  	  }         	  	
         	  } 
         	  
         	deviceWorkload[sel_no].workload = workload;         	  
          }      	
      }   
	
 }


class Workload  {
	
   /** ID of a candidate device. */
	String did;
	
	/** Amount of currently assigned workload on the device. */
	double workload;
	
}
