/*

* Copyright (c) 2005, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

import java.io.*;
import java.util.*;

/**
 * Implementation of the traditional greedy list scheduling (LS) algorithm. 
 *
 * @see SRFLS
 * @see LRFLS
 * @see AortaSchedulingSimulator#schedule
 * @see GroupingThread#scheduleCameras
 */ 
public class LS {
	
	Vector requests;
	
	Vector finishedAID = new Vector (20,10);
	
	PrintWriter logFileWriter;
	
	double overallServiceTime, overallSchedulingTime;
	
	CameraSimulator camSim [];
	
		
	/** 
	 * @param requests a vector of action requests to be scheduled
	 * @param logFileWriter A {@link PrintWriter} corresponds to the log file
	 * 
	 * @see ActionRequest
	 */
	 public LS (Vector requests, PrintWriter logFileWriter) 
	  {
		this.requests = requests;
		this.logFileWriter = logFileWriter;
		
		overallServiceTime = 0;  overallSchedulingTime = 0;	
      }  
    
   
    public void doAlgorithm () 
      {
    	int i, j,k; 
    	long startTime, endTime;
    	
    	startTime = System.currentTimeMillis();   	
    	  	
    	// decide the set of candidate devices
    	Vector didVector = new Vector (20,10);
    	for (i=0;i<requests.size();i++)
    	  for (j=0;j<((ActionRequest)requests.elementAt(i)).didVector.size();j++)
    	  	if (!didVector.contains((String)((ActionRequest)requests.elementAt(i)).didVector.elementAt(j)))
    	  	  didVector.addElement(((String)((ActionRequest)requests.elementAt(i)).didVector.elementAt(j)));
    	   	
    	
    	deviceRequestQueue devices [] = new deviceRequestQueue [didVector.size()];
    	camSim = new CameraSimulator [didVector.size()];
    	for (i=0;i<didVector.size();i++)
    	  {
    	  	camSim[i] = new CameraSimulator(String.valueOf(i+1));
    	  	
    	  	devices[i] = new deviceRequestQueue();
    	  	devices[i].did = (String)didVector.elementAt(i);
    	  	for (j=0;j<requests.size();j++)
    	  	  for (k=0;k<((ActionRequest)requests.elementAt(j)).didVector.size();k++)
    	  	    if ( ((String)((ActionRequest)requests.elementAt(j)).didVector.elementAt(k)).equals(devices[i].did) )
    	  	      {
    	  	        devices[i].requests.addElement((ActionRequest)requests.elementAt(j));
    	  	        break;
    	  	      } 	  	
    	  } 
    	  
    	endTime = System.currentTimeMillis();
    	
    	overallSchedulingTime += endTime - startTime;
    	
    	schedule(devices);  		
    }  
    
    
    /** Schedule the action requests to the candidate cameras. */
    public void schedule (deviceRequestQueue devices []) 
      {
    	int i,j;
    	int count = 0;
    	int finishedNum = 0;
    	int size = requests.size();
    	int id;
    	int pan, tilt, zoom;
    	int deltaPan, deltaTilt, deltaZoom;
    	long currentTime, startTime, endTime;
    	double connectionCost, headMovementCost, actionCost;        
        double singleServiceTime [] = new double [devices.length];
    	Tuple t;
    	ActionRequest ar;    	
    	
    	for (i=0;i<singleServiceTime.length;i++)
    	  singleServiceTime[i] = 0; 
    	  
    	connectionCost = 0;
 		for (i=0;i<camSim.length;i++)
 		  if (connectionCost < camSim[i].connection())
 		    connectionCost = camSim[i].connectionCost; 
 		overallSchedulingTime += connectionCost;	   	  
    	
    	logFileWriter.println("#");
    	while (finishedNum != size)
    	  { 
    	    currentTime = System.currentTimeMillis();
    	    
    	  	for (i=0;i<devices.length;i++)
    	  	  if (devices[i].availableTime < currentTime)  // the device is free
    	  	    {
    	  	      startTime = System.currentTimeMillis();
    	  	      for (j=0;j<devices[i].requests.size();j++)
    	  	        if (!finishedAID.contains(((ActionRequest)devices[i].requests.elementAt(j)).aid))  // unserviced request
    	  	          break;
    	  	   	      
    	  	      if (j == devices[i].requests.size())  // all requests that this device can service has been serviced
    	  	        {
    	  	          devices[i].availableTime = Double.MAX_VALUE;  // disable the device
    	  	          endTime = System.currentTimeMillis();
    	  	          overallSchedulingTime += endTime - startTime;
    	  	          continue;
    	  	        }
    	  	   	
    	  	      ar = (ActionRequest)devices[i].requests.elementAt(j);
    	  	      for (j=0;j<ar.didVector.size();j++)
    	  	        if (devices[i].did.equals((String)ar.didVector.elementAt(j)))
    	  	          {ar.sel_no = j; break;}
    	  	      id = Integer.parseInt(devices[i].did);
    	  	      t = (Tuple)ar.statusVector.elementAt(ar.sel_no);
    	  	  	  pan = Integer.parseInt((String)t.fields.elementAt(0));
    	   	      tilt = Integer.parseInt((String)t.fields.elementAt(1));
    	   	      zoom = Integer.parseInt((String)t.fields.elementAt(2));
    	   	      deltaPan = Math.abs(pan - camSim[id-1].pan);
    	   	      deltaTilt = Math.abs(tilt - camSim[id-1].tilt);
    	   	      deltaZoom = Math.abs(zoom - camSim[id-1].zoom);    	   	    
    	   	      headMovementCost = camSim[id-1].computeHeadMovementCost(deltaPan,deltaTilt, deltaZoom);
    	   	      actionCost = estimateActionCost(headMovementCost); 
    	   	  	  
    	   	  	  camSim[id-1].pan = pan;  	   	  
    	  	   	  camSim[id-1].tilt = tilt;
    	  	   	  camSim[id-1].zoom = zoom;
    	  	   	  
    	  	   	  finishedAID.addElement((String)ar.aid);
    	  	   	  finishedNum ++;
    	  	   	  
    	  	   	  endTime = System.currentTimeMillis();
    	  	   	  overallSchedulingTime += endTime - startTime;
    	  	   	  singleServiceTime[id-1] += actionCost;
    	  	   	  devices[i].availableTime = endTime + actionCost;
    	  	   	  
    	  	   	  logFileWriter.println("Round #" + (++count) + ": Request #" + ar.aid + " is serviced on Device " + id + ".");
    	  	      logFileWriter.println("Action cost: " + actionCost + " ms.");
    	  	      logFileWriter.println();    	  	      
    	  	      
    	  	      logFileWriter.println("Current physical status of Camera " + id + ": pan = " + pan + "  tilt = " + tilt + "  zoom = " + zoom + ".");
                  logFileWriter.println();   	  	        	  
    	  	    }                  	  	     
    	  }
    	   
        logFileWriter.println("#");
 	    for (i=0;i<singleServiceTime.length;i++)
 	      {
 	      	logFileWriter.println("Service time on Device " + (i+1) + ": " + singleServiceTime[i] + " ms.");
 	      	if (singleServiceTime[i] > overallServiceTime)
 	      	  overallServiceTime = singleServiceTime[i]; 
 	      	if (i == singleServiceTime.length-1)
 	      	  logFileWriter.println("#");
 	      	logFileWriter.println();	      	
 	      }  		
            
        logFileWriter.println("#");
 		logFileWriter.println("Service time: " + overallServiceTime + " ms.");
 		logFileWriter.println("Scheduling time: " + overallSchedulingTime + " ms.");
 		logFileWriter.println("Makespan: " + (overallServiceTime + overallSchedulingTime) + " ms.");
 		logFileWriter.println("#");  	
      } 
      
    /** Estimate the cost of a <code>photo()</code> action. */
    public double estimateActionCost (double headMovementCost)
      {
      	return 2*CameraSimulator.STANDARD_CONNECTION_TIME + headMovementCost + CameraSimulator.TAKE_MEDIUM_SIZE_PHOTO_ESTIMATED_COST;
      }    
	
 } 


class deviceRequestQueue 
  {
	/** ID of a candidate device. */
	String did;
	
    /** The set of requests that can be serviced on the device. */
	Vector requests = new Vector (30, 10);
	
	/** Indicate the time from which the device becomes free. */
	double availableTime = 0;
  } 


