/*

* Copyright (c) 2005, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

import java.io.*;
import java.util.*;

/**
 * Implementation of the random scheduling.
 *
 * @see AortaSchedulingSimulator#schedule
 * @see GroupingThread#scheduleCameras
 */
public class RANDOM {
	
	Vector requests;
	
	PrintWriter logFileWriter;
	
	deviceWorkload allDeviceWorkload [];
	
	double singleServiceTime, singleSchedulingTime;
 	double overallServiceTime, overallSchedulingTime;
 	

	/** 
	 * @param requests a vector of action requests to be scheduled
	 * @param logFileWriter A {@link PrintWriter} corresponds to the log file
	 *  
	 * @see ActionRequest
	 */
	public RANDOM (Vector requests, PrintWriter logFileWriter) 
	  {
		this.requests = requests;
		this.logFileWriter = logFileWriter;	
		
		overallServiceTime = 0; overallSchedulingTime = 0;
		
		long startTime = System.currentTimeMillis();
		
		initialize ();
		
		assignment();    	
      	
    	long endTime = System.currentTimeMillis();
    	
    	overallSchedulingTime += endTime - startTime;	
	  } 
	
	
	public void initialize ()
	  {
	  	int i,j;
	  	
	  	Vector didVector = new Vector (20,10);
    	for (i=0;i<requests.size();i++)
    	  for (j=0;j<((ActionRequest)requests.elementAt(i)).didVector.size();j++)
    	  	if (!didVector.contains((String)((ActionRequest)requests.elementAt(i)).didVector.elementAt(j)))
    	  	  didVector.addElement(((String)((ActionRequest)requests.elementAt(i)).didVector.elementAt(j)));
    
    	allDeviceWorkload = new deviceWorkload [didVector.size()];
    	
    	for (i=0;i<didVector.size();i++)
    	  {
    	  	allDeviceWorkload[i] = new deviceWorkload();
    	  	allDeviceWorkload[i].did = (String)didVector.elementAt(i);
    	  }     	
	  }
	  
	  
	/** Assign the requests to the candidate cameras in a random fashion. */
	public void assignment ()
	  {
	   	int i,j;
		int sel_no;
		
		Random ran = new Random();
		ActionRequest ar;
		
		// assign the requests to devices in a pure random fashion
    	for (i=0;i<requests.size();i++)
    	  {
    	   	ar = (ActionRequest)requests.elementAt(i);
    	   	
    	   	sel_no = Math.abs(ran.nextInt()) % ar.didVector.size();
    	  	
    	  	String did = (String)ar.didVector.elementAt(sel_no);
    	  	
    	  	for (j=0;j<allDeviceWorkload.length;j++)
    	  	  if (allDeviceWorkload[j].did.equals(did))
    	  	    break;
    	  	    
    	  	allDeviceWorkload[j].requests.addElement((ActionRequest)ar);
    	  	
    	  	ar.sel_no = sel_no;    	  	
    	  }	
	  }
	  
	  
	/** Execute the requests on the candidate cameras. */
	public void execution() 
	  {
		int i;
		double maxSingleServiceTime = 0, maxSingleSchedulingTime = 0;
		 	  	
 	  	for (i=0;i<allDeviceWorkload.length;i++)
 	  	  {
 	  	  	logFileWriter.println("\n#");
 	  	  	logFileWriter.println("Device " + allDeviceWorkload[i].did + " :");
 	  	  	singleDeviceExecution(allDeviceWorkload[i]);
 	  	  	if (singleServiceTime > maxSingleServiceTime) 
 	  	  	  maxSingleServiceTime = singleServiceTime;
 	  	  	if (singleSchedulingTime > maxSingleSchedulingTime)
 	  	  	  maxSingleSchedulingTime = singleSchedulingTime;
 	  	  	logFileWriter.println("#"); 	  	  	 	  	
 	      } 
 	    
 	    overallServiceTime += maxSingleServiceTime; 
 	    overallSchedulingTime += maxSingleSchedulingTime;  
 	    
 	    logFileWriter.println();
 	    logFileWriter.println("#");
 	    logFileWriter.println("Service time: " + overallServiceTime + " ms.");
 	    logFileWriter.println("Scheduling time: " + overallSchedulingTime + " ms.");
 	    logFileWriter.println("Makespan: " + (overallServiceTime + overallSchedulingTime) + " ms.");
 	    logFileWriter.println("#");	    		
	  } 
	  
	
	/** Execute the requests that have been assigned to a single camera in a random fasihon. */
	public void singleDeviceExecution(deviceWorkload dw)  
      {
	    int sel_no;
    	int count = 0;
    	int pan, tilt, zoom;
    	int deltaPan, deltaTilt, deltaZoom;
    	long startTime, endTime;
    	double headMovementCost, actionCost;
    	Random ran = new Random(); 	
    	Tuple t;
    	ActionRequest ar;
    	CameraSimulator camSim;    	
    	
    	camSim = new CameraSimulator(dw.did);    	
    	singleServiceTime = 0; singleSchedulingTime = 0;
    	singleSchedulingTime += camSim.connection();    	
    	while (dw.requests.size() != 0)
    	  {
    	  	logFileWriter.println("Current physical status of Camera " + camSim.id + ": pan = " + camSim.pan + "  tilt = " + camSim.tilt + "  zoom = " + camSim.zoom + ".");
            logFileWriter.println();
             
            startTime = System.currentTimeMillis(); 
               	  	
    	  	if (dw.requests.size() == 1)
    	  	  sel_no = 0;
    	  	else
    	  	  sel_no = Math.abs(ran.nextInt()) % dw.requests.size();    	  	
    	   	
    	  	endTime = System.currentTimeMillis();
    	  	singleSchedulingTime += endTime - startTime;
    	  	    	  	
    	  	ar = (ActionRequest)dw.requests.elementAt(sel_no);
    	  	t = (Tuple)ar.statusVector.elementAt(ar.sel_no);
    	  	pan = Integer.parseInt((String)t.fields.elementAt(0));
    	   	tilt = Integer.parseInt((String)t.fields.elementAt(1));
    	   	zoom = Integer.parseInt((String)t.fields.elementAt(2));
    	   	deltaPan = Math.abs(pan - camSim.pan);
    	   	deltaTilt = Math.abs(tilt - camSim.tilt);
    	   	deltaZoom = Math.abs(zoom - camSim.zoom);
    	   	headMovementCost = camSim.computeHeadMovementCost(deltaPan, deltaTilt, deltaZoom);
    	   	actionCost = estimateActionCost(headMovementCost); 
    	   	singleServiceTime += actionCost; 
    	   	
    	  	camSim.pan = Integer.parseInt((String)t.fields.elementAt(0));
    	   	camSim.tilt = Integer.parseInt((String)t.fields.elementAt(1));
    	   	camSim.zoom = Integer.parseInt((String)t.fields.elementAt(2)); 
    	  	
    	  	logFileWriter.println("Round #" + (++count) + ": Request #" + ar.aid + " is serviced. ");
    	  	logFileWriter.println("Action cost: " + actionCost + " ms.");
    	  	logFileWriter.println();
    	  	    	  	 
    	  	dw.requests.removeElementAt(sel_no);   	  	   	  	
    	  }    	
    	 
    	logFileWriter.println("Service time on this device: " + singleServiceTime + " ms.");
 	  	logFileWriter.println("Scheduling time on this device: " + singleSchedulingTime + " ms."); 	  	 	  	  	  	     	
      } 
       
      
    /** Estimate the cost of a <code>photo()</code> action. */
    public double estimateActionCost (double headMovementCost)
      {
      	return 2*CameraSimulator.STANDARD_CONNECTION_TIME + headMovementCost + CameraSimulator.TAKE_MEDIUM_SIZE_PHOTO_ESTIMATED_COST;
      }  
	
 } 